import os
from collections import namedtuple

from astropy import table

FileRecord = namedtuple(
    "FileRecord",
    ["filepath", "db", "comment", "existence"]
)


class FileRecorder(list):
    """
    FileRecord Recorder, inherited from the built-in ``list``.

    This recorder is used to record files generated by functional modules.
    In principle, a CSST pipeline module should return a status (CsstStatus)
    and a file recorder (``FileRecorder``) after it finishes data processing.

    Parameters
    ----------
    *args : any
        The arguments passed to ``list()``.
    **kwargs : any
        The keyword arguments passed to ``list()``.

    Methods
    -------
    add_record(filepath: str = "", db: bool = False, comment: str = "")
        Add a file record, each record is a ``collections.namedtuple``.
        Users should provide a file path ``filepath``, whether this file should be
        written to database ``db``, a comment ``comment``.
        An existence boolean will be attached to check if the file exists.
    to_table()
        Convert to ``astropy.table.Table``.
    pprint(*args, **kwargs)
        Use ``astropy.table.Table.pprint`` to print file records in table format.
    pprint_all(*args, **kwargs)
        Use ``astropy.table.Table.pprint_all`` to print file records in table format.

    Examples
    --------
    >>> fr = FileRecorder()
    >>> for i in range(3):
    >>>     fr.add_record("test{:03d}.txt".format(i), db=True, comment="Test file {:d}".format(i))
    >>> fr.pprint_all()
    <FileRecorder length=3>
      filepath   db    comment   existence
    ----------- ---- ----------- ---------
    test000.txt True Test file 0     False
    test001.txt True Test file 1     False
    test002.txt True Test file 2     False
    >>> fr.pprint_all()
    <FileRecorder length=3>
      filepath   db    comment   existence
    ----------- ---- ----------- ---------
    test000.txt True Test file 0     False
    test001.txt True Test file 1     False
    test002.txt True Test file 2     False
    """

    def __init__(self, *args, **kwargs):
        super(FileRecorder, self).__init__(*args, **kwargs)

    @staticmethod
    def FileRecord(filepath: str = "", db: bool = False, comment: str = ""):
        return FileRecord(filepath=filepath, db=db, comment=comment, existence=os.path.exists(filepath))

    def add_record(self, filepath: str = "", db: bool = False, comment: str = ""):
        existence = os.path.exists(filepath)
        assert isinstance(filepath, str)
        assert isinstance(db, bool)
        assert isinstance(comment, str)
        super().append(FileRecord(filepath=filepath, db=db, comment=comment, existence=existence))

    def to_table(self):
        return table.Table([_._asdict() for _ in self])

    def pprint(self, *args, **kwargs):
        print("<FileRecorder length={}>".format(len(self)))
        return self.to_table().pprint(*args, **kwargs)

    def pprint_all(self, *args, **kwargs):
        print("<FileRecorder length={}>".format(len(self)))
        return self.to_table().pprint_all(*args, **kwargs)

    def __repr__(self):
        t = self.to_table()
        lines, outs = t.formatter._pformat_table(
            t, max_lines=-1, max_width=-1, show_name=True, show_unit=None, show_dtype=False, align="<")
        if outs['show_length']:
            lines.append(f'Length = {len(self)} rows')
        return "\n".join(lines)

    @property
    def summary(self):
        return f"{len(self)} alleged, " \
               f"{len(self) - sum(self.to_table()['existence'])} missing, " \
               f"{sum(self.to_table()['existence'])} to db"
